@@ -1,10 +1,16 @@
-#
-# Debatable flask-side auth management, and more.
-#
-# This code is public domain.
-#
+"""
+Flask Extension and Wrapper
 
-from typing import Optional, Callable, Dict, Any
+This extension helps manage:
+- authentication
+- authorization
+- parameters
+- and more…
+
+This code is public domain.
+"""
+
+from typing import Optional, Callable, Dict, List, Any
 
 import functools
 import inspect
@@ -21,30 +27,38 @@ import logging
 log = logging.getLogger("fsa")
 
 
-# carry data for error Response
 class AuthException(BaseException):
+    """Exception class to carry fields for an error Response."""
+
     def __init__(self, message: str, status: int):
+        """Constructor parameters:
+
+        - message: Response's message
+        - status: intended HTTP status
+        """
         self.message = message
         self.status = status
 
 
-#
 # special type casts
 #
 def bool_cast(s: str) -> Optional[bool]:
+    """Parses a bool."""
     return None if s is None else \
         False if s.lower() in ("", "0", "false", "f") else \
         True
 
 
 def int_cast(s: str) -> Optional[int]:
+    """Parses an integer, allowing several bases."""
     return int(s, base=0) if s is not None else None
 
 
-# special "path" type
 class path(str):
+    """Type to distinguish path parameters."""
     pass
 
+
 # note: mypy complains wrongly about non-existing _empty.
 CASTS: Dict[type, Callable[[str], object]] = {
     bool: bool_cast,
@@ -62,8 +76,8 @@ ALL = "ALL authentified users are allowed"
 NONE = "NONE can come in, the path is forbidden"
 
 
-# guess parameter type
 def typeof(p: inspect.Parameter):
+    """Guess parameter type, possibly with some type inference."""
     if p.kind == p.VAR_KEYWORD:
         return dict
     elif p.kind == p.VAR_POSITIONAL:
@@ -76,18 +90,34 @@ def typeof(p: inspect.Parameter):
         return str
 
 
-# convenient object wrapper class
 class Reference:  # type: Any
+    """Convenient object wrapper class.
 
-    def __init__(self, obj: Any = None):
+    The wrapper forwards most method calls to the wrapped object.
+    """
+
+    def __init__(self, obj: Any = None, set_name: str = "set"):
+        """Constructor parameters:
+
+        - obj: object to be wrapped, can also be provided later.
+        - set_name: provide another name for the "set" function.
+        """
         self._obj = None
-        # keep track of initial methods
-        self._init = set(self.__dir__())
-        self._init.add("_init")
+        # possibly rename the "set" method
+        if set_name is None:
+            set_name = "set"
+        elif set_name != "set":
+            setattr(self, set_name, getattr(self, "set"))
+            # cannot really delete "set"…
+            # delattr(self, "set"): attribute error
+            # delattr(self.__class__, "set"): removes for all instances!
+        # keep track of initial methods for later cleanup
+        self._init = set(self.__dir__() + ["_init"])
         if obj is not None:
-            self._setobj(obj)
+            getattr(self, set_name)(obj)
 
-    def _setobj(self, obj):
+    def set(self, obj):
+        """Set current wrapped object, possibly replacing the previous one."""
         log.debug(f"setting reference to {obj} ({type(obj)})")
         self._obj = obj
         # method cleanup
@@ -98,39 +128,50 @@ class Reference:  # type: Any
         for f in obj.__dir__():
             if f not in self._init:
                 setattr(self, f, getattr(obj, f))
-        # and more forwards
-        # for f in ("str", "repr", "hash", "eq", "ne", "lt", "gt", "le", "ge"):
-        #    name = "__" + f + "__"
-        #    setattr(self, name, getattr(obj, name))
 
+    # forward standard methods
+    # automating that with setattr/getattr does not work…
     def __str__(self):
         return self._obj.__str__()
 
     def __repr__(self):
         return self._obj.__repr__()
 
-    def __eq__(self, *args):
-        return self._obj.__eq__(*args)
-
     def __hash__(self):
         return self._obj.__hash__()
 
+    def __eq__(self, o):
+        return self._obj.__eq__(o)
+
+    def __ne__(self, o):
+        return self._obj.__ne__(o)
+
+    def __le__(self, o):
+        return self._obj.__le__(o)
+
+    def __lt__(self, o):
+        return self._obj.__lt__(o)
+
+    def __ge__(self, o):
+        return self._obj.__ge__(o)
+
+    def __gt__(self, o):
+        return self._obj.__gt__(o)
+
 
-#
-# Positive caching.
-#
-# Keep a cache of true answers to reduce database queries,
-# but keep querying on failures.
-#
-# See also: functools.cache, functools.lru_cache
 #
 # TODO
 # - LRU? LFU?
 # - automatic reset based on cache efficiency? expansion?
 #
 class CacheOK:
+    """Positive caching decorator.
 
-    def __init__(self, fun):
+    Cache True answers, but still forwards False answers to the underlying
+    function.
+    """
+
+    def __init__(self, fun: Callable[[List[Any]], bool]):
         self._fun = fun
         self._cache = set()
         self.cache_clear = self._cache.clear
@@ -145,8 +186,17 @@ class CacheOK:
             return ok
 
 
-# Flask wrapper
 class Flask(flask.Flask):
+    """Flask class wrapper.
+
+    The class behaves mostly as a Flask class, but supports extensions:
+
+    - the `route` decorator manages authentication, authorization and
+      parameters transparently.
+    - several additional methods are provided: `init_app`, `get_user_pass`,
+      `user_in_group`, `check_password`, `hash_password`, `create_token`,
+      `get_user`, `current_user`.
+    """
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -156,44 +206,55 @@ class Flask(flask.Flask):
         self.add_url_rule = self._fsa.add_url_rule
 
     # forward some methods
-    def init_app(self, app: flask.Flask):
-        return self._fsa.init_app(app)
+    def init_app(self):
+        """Initialize underlying extension."""
+        return self._fsa.init_app(self)
 
     def get_user_pass(self, gup):
+        """Set `get_user_pass` helper function."""
         return self._fsa.get_user_pass(gup)
 
     def user_in_group(self, uig):
+        """Set `user_in_group` helper function."""
         return self._fsa.user_in_group(uig)
 
     def check_password(self, pwd, ref):
+        """Check whether password is ok wrt to current configuration."""
         return self._fsa.check_password(pwd, ref)
 
     def hash_password(self, pwd):
+        """Hash password using current password scheme."""
         return self._fsa.hash_password(pwd)
 
     def create_token(self, user):
+        """Create a token with the current token configuration."""
         return self._fsa.create_token(user)
 
     def get_user(self):
+        """Authenticate remote user."""
         return self._fsa.get_user()
 
     def current_user(self):
-        return self._fsa._user
+        """Get current authenticated user, if any."""
+        return self._fsa.current_user()
 
 
 # actual class
 class FlaskSimpleAuth:
+    """Flask extension to implement authentication, authorization and parameter
+    management.
+    """
 
-    # constructor
     def __init__(self, app: flask.Flask = None):
+        """Constructor parameter: flask application to extend."""
         self._app = app
         self._get_user_pass = None
         self._user_in_group = None
         # actual initialization is delayed
         self._initialized = False
 
-    # set, or possibly just reset, the current authentication
     def _auth_set_user(self):
+        """Before request hook to perform early authentication."""
         self._user = None
         self._need_authorization = True
         if not self._always:
@@ -207,8 +268,9 @@ class FlaskSimpleAuth:
             return e.message, e.status
         assert self._user is not None
 
-    # wipe out current authentication
     def _auth_after_cleanup(self, res: Response):
+        """After request hook to cleanup authentication and detect missing
+        authorization."""
         self._user = None
         if res.status_code < 400 and self._need_authorization:
             method, path = request.method, request.path
@@ -217,22 +279,27 @@ class FlaskSimpleAuth:
                 return Response("missing authorization check", 500)
         return res
 
-    # store get_user_pass helper function, can be used as a decorator
     def get_user_pass(self, gup):
+        """Set `get_user_pass` helper, can be used as a decorator."""
         self._get_user_pass = gup
         return gup
 
-    # store user_in_group helper function, can be used as a decorator
     def user_in_group(self, uig):
+        """Set `user_in_group` helper, can be used as a decorator."""
         self._user_in_group = uig
         return uig
 
     def initialize(self):
+        """Run late initialization on current app."""
         assert self._app is not None
         self.init_app(self._app)
 
-    # actually initialize class…
     def init_app(self, app: flask.Flask):
+        """Initialize extension with a Flask application.
+
+        The initialization is performed through `FSA_*` configuration
+        directives.
+        """
         log.info("FSA initialization…")
         assert app is not None
         self._app = app
@@ -293,7 +360,7 @@ class FlaskSimpleAuth:
             elif algo == "none":
                 self._sign = None
             else:
-                raise Exception("unexpected jwt FSA_TOKEN_ALGO ({algo})")
+                raise Exception(f"unexpected jwt FSA_TOKEN_ALGO ({algo})")
             self._siglen = 0
         else:
             raise Exception(f"invalid FSA_TOKEN_TYPE ({self._type})")
@@ -361,12 +428,12 @@ class FlaskSimpleAuth:
     # note: passlib bcrypt is Apache compatible
     #
 
-    # verify password
     def check_password(self, pwd, ref):
+        """Verify whether a password is correct."""
         return self._pm.verify(pwd, ref)
 
-    # hash password consistently with above check, can be used by app
     def hash_password(self, pwd):
+        """Hash password according to the current password scheme."""
         return self._pm.hash(pwd)
 
     # check user password against internal credentials
@@ -465,8 +532,8 @@ class FlaskSimpleAuth:
         return jwt.encode({"exp": exp, "sub": user, "aud": realm},
                           secret, algorithm=self._algo)
 
-    # create a new token for user depending on the configuration
     def create_token(self, user):
+        """Create a new token for user depending on the configuration."""
         assert self._type is not None
         realm, delay = self._realm, self._delay
         if self._type == "fsa":
@@ -541,8 +608,8 @@ class FlaskSimpleAuth:
         "fake": _get_fake_auth
     }
 
-    # return authenticated user or throw exception
     def get_user(self):
+        """Authenticate user or throw exception."""
         log.debug(f"get_user for {self._auth}")
 
         # _user is reset before/after requests
@@ -588,6 +655,10 @@ class FlaskSimpleAuth:
         log.info(f"get_user({self._auth}): {self._user}")
         return self._user
 
+    def current_user(self):
+        """Return current authenticated user, if any."""
+        return self._user
+
     #
     # authorize internal decorator
     #
@@ -721,11 +792,10 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    #
-    # route decorator helper
-    #
     def add_url_rule(self, rule, endpoint=None, view_func=None, authorize=NONE, required=None, allparams=False, **options):
 
+        """Route decorator helper method."""
+
         # lazy initialization
         if not self._initialized:
             self.init_app(self._app)
@@ -764,12 +834,13 @@ class FlaskSimpleAuth:
         aut = self._authorize(*roles)(par)
         return flask.Flask.add_url_rule(self._app, newpath, endpoint=endpoint, view_func=aut, **options)
 
-    # route decorator
     def route(self, rule, **options):
+        """Extended `route` decorator provided by the extension."""
         def decorate(fun):
             self.add_url_rule(rule, view_func=fun, **options)
         return decorate
 
     # duck-typing blueprint code stealing: needs blueprints, _blueprint_order, debug
     def register_blueprint(self, blueprint, **options):
+        """Register a blueprint."""
         flask.Flask.register_blueprint(self, blueprint, **options)
